class_name Board
extends Node2D

const _scn: PackedScene = preload("res://src/complete_board.tscn")
const launch_script = preload('res://src/launch_zone.gd')

## Gates parameters
# var col0 = [150.0, 240.0, 330.0]
# var col1 = [110.0, 195.0, 285.0, 370.0]
const x_shift = -20 # marble diameter 16 + 4 to be large
var col0 = [
	launch_script.x0 + 2*launch_script.x_increment + x_shift,
	launch_script.x0 + 4*launch_script.x_increment + x_shift,
	launch_script.x0 + 6*launch_script.x_increment + x_shift,
 ]
var col1 = [
	launch_script.x0 + 1*launch_script.x_increment + x_shift,
	launch_script.x0 + 3*launch_script.x_increment + x_shift,
	launch_script.x0 + 5*launch_script.x_increment + x_shift,
	launch_script.x0 + 7*launch_script.x_increment + x_shift,
 ]

var y_increment = 135.0
var row0_y = 295.0 
var row1_y = row0_y + y_increment
var row2_y = row0_y + y_increment * 2
var row3_y = row0_y + y_increment * 3

var setup_test0 = [
	[0,0,0],
	[0,0,0,0],
	[0,0,0,0],
	[0,0,0,0]
]


# Called when the node enters the scene tree for the first time.
func _ready() -> void:
	setup_internal_walls()
	
func setup_puzzle(angles, marbles, x_shift_puzzle=0.0, y_shift_puzzle=0.0):
	AudioServer.set_bus_mute(3, true)
	var marble_refs = []
	var gate = load("res://src/gate.tscn")
	var marble_scn = load("res://src/marble.tscn")
	var vector_shift = Vector2(x_shift_puzzle, y_shift_puzzle)
	var row0 = [Vector2(col0[0],row0_y), Vector2(col0[1],row0_y), Vector2(col0[2],row0_y)] 
	var row1 = [Vector2(col1[0],row1_y), Vector2(col1[1],row1_y), Vector2(col1[2],row1_y), Vector2(col1[3],row1_y)] 
	var row2 = [Vector2(col0[0],row2_y), Vector2(col0[1],row2_y), Vector2(col0[2],row2_y)] 
	var row3 = [Vector2(col1[0],row3_y), Vector2(col1[1],row3_y), Vector2(col1[2],row3_y), Vector2(col1[3],row3_y)] 
	var gates = [row0, row1, row2, row3]
	for i in len(gates):
		var row = gates[i]
		for j in len(row):
			var comp_position = row[j] + vector_shift
			var new_gate = gate.instantiate()
			new_gate.name = "Gate_" + str(i) + "_" + str(j)
			new_gate.set_position(comp_position)
			new_gate.initial_position = comp_position
			var row_angles = angles[i]
			var setup_angle = row_angles[j] 
			new_gate.set_angle(setup_angle)
			add_child(new_gate)
			if setup_angle != 0:
				var row_marbles = marbles[i]
				var current_marble = row_marbles[j]
				if current_marble != null:
					var new_marble = marble_scn.instantiate()
					new_marble.set_color(current_marble)
					var y = comp_position.y - 70 
					var x = comp_position.x + 20*setup_angle 
					new_marble.position = Vector2(x,y)
					new_marble.name = "MarblePuzzle_" + str(i) + "_" + str(j)
					marble_refs.append(new_marble)
					add_sibling(new_marble)
	return marble_refs

func setup_board(board_initial_state) -> Board:
	var result: Board = _scn.instantiate()
	result.setup_gates(false, board_initial_state)
	return result
	
func setup_random_board() -> Board:
	var result: Board = _scn.instantiate()
	result.setup_gates(true, null)
	return result
	
func setup_internal_walls():
	const vertical_shift = -30
	var int_wall_y_row_0 = row0_y + vertical_shift
	var int_wall_y_row_1 = row1_y + vertical_shift
	var int_wall_y_row_2 = row2_y + vertical_shift
	var int_wall_y_row_3 = row3_y + vertical_shift
	var internal_wall = load("res://src/internal_wall.tscn")
	var int_wall_col0_0 = (col0[0]+col0[1])/2.
	var int_wall_col0_1 = (col0[1]+col0[2])/2.
	var row0 = [Vector2(int_wall_col0_0,int_wall_y_row_0), Vector2(int_wall_col0_1,int_wall_y_row_0)] 
	var int_wall_col1_0 = (col1[0]+col1[1])/2.
	var int_wall_col1_1 = (col1[1]+col1[2])/2.
	var int_wall_col1_2 = (col1[2]+col1[3])/2.
	var row1 = [Vector2(int_wall_col1_0,int_wall_y_row_1), Vector2(int_wall_col1_1,int_wall_y_row_1), Vector2(int_wall_col1_2,int_wall_y_row_1)] 
	var int_wall_col2_1 = (col0[0]+col0[1])/2.
	var int_wall_col2_2 = (col0[1]+col0[2])/2.
	var int_wall_col2_0 = col0[0] - (col0[1]-col0[0])/2.
	var int_wall_col2_3 = col0[2] + (col0[2]-col0[1])/2.
	var row2 = [Vector2(int_wall_col2_0,int_wall_y_row_2), Vector2(int_wall_col2_1,int_wall_y_row_2), Vector2(int_wall_col2_2,int_wall_y_row_2), Vector2(int_wall_col2_3,int_wall_y_row_2)] 
	var row3 = [Vector2(int_wall_col1_0,int_wall_y_row_3), Vector2(int_wall_col1_1,int_wall_y_row_3), Vector2(int_wall_col1_2,int_wall_y_row_3)] 
	
	var internal_walls = [row0, row1, row2, row3]
	for i in len(internal_walls):
		var row = internal_walls[i]
		for j in len(row):
			var wall_position = row[j]
			var new_wall = internal_wall.instantiate()
			new_wall.name = "iWall_" + str(i) + "_" + str(j)
			new_wall.set_position(wall_position)
			add_child(new_wall)

func setup_gates(random, angles):
	AudioServer.set_bus_mute(3, true)
	var gate = load("res://src/gate.tscn")
	var row0 = [Vector2(col0[0],row0_y), Vector2(col0[1],row0_y), Vector2(col0[2],row0_y)] 
	var row1 = [Vector2(col1[0],row1_y), Vector2(col1[1],row1_y), Vector2(col1[2],row1_y), Vector2(col1[3],row1_y)] 
	var row2 = [Vector2(col0[0],row2_y), Vector2(col0[1],row2_y), Vector2(col0[2],row2_y)] 
	var row3 = [Vector2(col1[0],row3_y), Vector2(col1[1],row3_y), Vector2(col1[2],row3_y), Vector2(col1[3],row3_y)] 
	var gates = [row0, row1, row2, row3]
	for i in len(gates):
		var row = gates[i]
		for j in len(row):
			var comp_position = row[j]
			var new_gate = gate.instantiate()
			new_gate.name = "Gate_" + str(i) + "_" + str(j)
			new_gate.set_position(comp_position)
			new_gate.initial_position = comp_position
			if random:
				new_gate.set_angle(2*(randi() % 2) - 1)
			else:
				var row_angles = angles[i]
				var setup_angle = row_angles[j] 
				new_gate.set_angle(setup_angle)
			add_child(new_gate)
	

func clean_all():
	# Use in tests to re-setup
	for node in get_children():
		if node.name != "BackBoard" and node.is_class("RigidBody2D"):
			remove_child(node)
			node.queue_free()

func read_gates() -> Array:
	var gate_gd = load("res://src/gate.gd")
	var gates = [
		  [0,0,0],
		[0,0,0,0],
		  [0,0,0],
		[0,0,0,0]
	]
	var read_degree = int(gate_gd.init_degree/2) + 1
	for node in get_children():
		var name_in_loop = String(node.name)
		if name_in_loop.begins_with("Gate_"):
			@warning_ignore("integer_division")
			var gate_lvl = int(name_in_loop)/10 
			var gate_pos = int(name_in_loop)%10  
			var log_ang = int(node.rotation_degrees/read_degree)
			gates[gate_lvl][gate_pos] = log_ang
			
	return gates

func read_marbles() -> Array:
	var marble_gd = preload('res://src/marble.gd')
	var color_dico = {
		"red": marble_gd.red,
		"green": marble_gd.green,
		"yellow": marble_gd.yellow }
	var marbles = [
		   [null, null, null],
		[null, null, null, null],
		   [null, null, null],
		[null, null, null, null]
	]
	for node in $BackBoard.get_children():
		var name_in_loop = String(node.name)
		if name_in_loop.begins_with("GateSection"):
			@warning_ignore("integer_division")
			var gate_lvl = int(name_in_loop)/10 - 1
			var gate_pos = int(name_in_loop)%10 - 1 
			if node.marbles.size() == 1:
				marbles[gate_lvl][gate_pos] = color_dico[node.marbles[0].get_color()]
			elif node.marbles.size() > 1:
				marbles[gate_lvl][gate_pos] = -1 # TODO Raise an error
	return marbles
