extends Sprite2D
class_name FancyBoardScore

const black = Color(0.0, 0.0, 0.0)
const lightgrey = Color(0.8, 0.8, 0.8)
const darkblue = Color(0.4, 0.4, 0.9)
const alpha = Color(0.4, 0.4, 0.9, 0.0)
const red = Color(0.82, 0.2, 0.2)
const blue = Color(0.2, 0.2, 0.82)
const yellow = Color(0.82, 0.82, 0.2)
const green = Color(0.2, 0.7, 0.2)
const colors = [green, red, yellow]
const color_names = ["green", "red", "yellow"]
const VictoryScore:int = 9 

var marble_scene = preload("res://src/marble_skin.tscn")
var marble_animation = preload("res://src/marble_animation.tscn")
var marble_shape = preload("res://src/marble_shape.tscn")

var target_marbles: Array

func check_score(prev_score)-> Dictionary:
	var main = get_parent()
	var puntos:int = prev_score
	
	var remaining_marbles = []
	for marble in main.get_children():
		if is_instance_valid(marble) and marble is Marble and marble.felt_to_bottom:
			remaining_marbles.append(marble)
	var current_marbles = remaining_marbles
	var increased_points = true #initiate as true to enter the while loop
	
	while increased_points and puntos < VictoryScore:
		increased_points = false
		## It loops over current_marbles to avoid changing array while looping over it
		current_marbles = remaining_marbles
		remaining_marbles = []
		for marble in current_marbles:
			if puntos >= VictoryScore:
				return {
					"points" : puntos,
					"marbles" : remaining_marbles
					}
			if marble.get_color() != color_names[self.target_marbles[puntos]]:
				remaining_marbles.append(marble)
			else:
				var initial_position = marble.position
				main.remove_child(marble)
				marble.queue_free()
				@warning_ignore("integer_division")
				var j = round(puntos / 3)
				var i = puntos - j*3
				var magic = marble_animation.instantiate()
				magic.z_index = self.z_index
				var new_point = magic.get_child(0) #Skin
				var marble_animation_effect = magic.get_child(1) #Effect
				var animatron = new_point.get_child(0)
				var anime = animatron.get_animation("scoring")
				anime.track_set_key_value(0,0, initial_position)

				# TODO : Change the 0.8 magic number for scale
				var rel_x = (i*80 - 80) * self.scale.x
				var rel_y = (j*80 - 80) * self.scale.y
				var new_position = self.position + Vector2(rel_x, rel_y) 
				anime.track_set_key_value(0,1, new_position)
				marble_animation_effect.position = new_position 
				new_point.set_color(self.target_marbles[i + j*3])
				main.add_child(magic) # Easier to get the position reference
				animatron.play("scoring")
				await animatron.animation_finished
				puntos += 1
				increased_points = true

	return {
		"points" : puntos,
		"marbles" : remaining_marbles
		}
	
func calculate_score(prev_score, marbles_indexes)-> int:
	var puntos:int = prev_score
	
	var remaining_marbles = marbles_indexes.duplicate(true)  # Deep copy
	var current_marbles = null
	var increased_points = true #initiate as true to enter the while loop
	
	while increased_points and puntos < VictoryScore:
		increased_points = false
		## It loops over current_marbles to avoid changing array while looping over it
		current_marbles = remaining_marbles.duplicate(true)  # Deep copy
		remaining_marbles = []
		for marble in current_marbles:
			if puntos >= VictoryScore:
				return puntos
			if marble != self.target_marbles[puntos]:
				remaining_marbles.append(marble)
			else:
				puntos += 1
				increased_points = true
				
	return puntos


func _ready():
	setup()
	
func setup():
	# Set cardboard target_marbles random
	randomize()
	for k in range(100): # Maximum number of tries
		target_marbles = []
		for i in range(9):
			target_marbles.append(randi() % 3)
		if check_entropy():
			break

func check_entropy() -> bool:
	## It validate the randomness
	var green_in_goal = Marble.green in target_marbles
	var red_in_goal = Marble.red in target_marbles
	var yellow_in_goal = Marble.yellow in target_marbles
	var there_are_three_colors = green_in_goal and red_in_goal and yellow_in_goal
	
	var three_in_a_row = false
	for i in range(7):
		if target_marbles[i] == target_marbles[i+1] and target_marbles[i] == target_marbles[i+2]:
			three_in_a_row = true
			
	return there_are_three_colors and not three_in_a_row

func _draw():
	# Draw 9 circles	
	for i in range(3):
		for j in range(3):
			var new_mark = marble_shape.instantiate()
			# TODO : Change the 0.8 magic number for scale
			var p0 = Vector2((i*100 - 100)*0.8, (j*100 - 100)*0.8)
			draw_circle_arc(p0, 28.0*0.8, 0, 360, colors[target_marbles[i + j*3]], 2)
			draw_circle_arc(p0, 26.0*0.8, 0, 360, colors[target_marbles[i + j*3]], 2)
			draw_circle_arc(p0, 24.0*0.8, 0, 360, colors[target_marbles[i + j*3]], 2)
			draw_circle_arc(p0, 30.0*0.8, 0, 360, black, 2)
				# TODO : Change the 80 magic number for scale
			var rel_x = 100*(i - 1) * self.scale.x
			var rel_y = 100*(j - 1) * self.scale.y
			#new_mark.position = self.position + Vector2(rel_x, rel_y) 
			new_mark.position = Vector2(rel_x, rel_y) 
			new_mark.set_color(target_marbles[i + j*3])
			self.add_child(new_mark)
			new_mark.z_as_relative = true
			new_mark.z_index = -1


func draw_circle_arc(center, radius, angle_from, angle_to, color, width):
	var nb_points = 32
	var points_arc = PackedVector2Array()

	for i in range(nb_points + 1):
		var angle_point = deg_to_rad(angle_from + i * (angle_to-angle_from) / nb_points - 90)
		points_arc.push_back(center + Vector2(cos(angle_point), sin(angle_point)) * radius)


	for index_point in range(nb_points):
		draw_line(points_arc[index_point], points_arc[index_point + 1], color, width)
