extends Node2D

signal gotopuzzle(puzzle_name) # Necessary for puzzle switcher don't crash
var solver_src = preload('res://src/solo_mode/analytic_solver.gd')
signal end_turn
var last_puzzle = 12
@export var puzzle_number:int = -1
@export var setup_gates:Array = [
	  [0,0,0],
	 [0,0,0,0],
	  [0,0,0],
	 [0,0,0,0]
]
@export var setup_marbles:Array = [
		   [null, null, null],
		[null, null, null, null],
		   [null, null, null],
		[null, null, null, null]
]

@export var puzzle_goal : Array = [-1, -1, -1]
@export var open_gate_condition : bool = false

var starting_time :int
var turn = 0
var turn_running = false
var launch_input_choice = null
var counting_stop_turn = 0
var basket_marbles_color = []

const threshold = 760. 

# Called when the node enters the scene tree for the first time.
func _ready() -> void:
	TranslationServer.set_locale($"/root/SettingsData".lang)
	$BotonConfirm.hide()
	$"/root/BackgroundMusic".pause_menu_music()
	$"/root/BackgroundMusic".pause_puzzle_music()
	$"/root/BackgroundMusic".play_puzzle_music()
	var validation = find_child("Validation") 
	validation.body_entered.connect(_on_marble_entering_goal)
	end_turn.connect(end_puzzle)
	%Title.text = tr("PZZLE_N") + " " + str(puzzle_number)
	%Target.text = tr("ScOre")
	%Condition.text = "[center]%s[/center]" % tr("N0OpenPath")
	if open_gate_condition:
		%Condition.visible = true
		%ConditionBG.visible = true
	else:
		%Condition.visible = false
		%ConditionBG.visible = false
	$Board.setup_puzzle(setup_gates, setup_marbles, $Board.position.x, $Board.position.y)
	for i in range(3):
		var target_marble = find_child("TargetMarble"+str(i))
		if puzzle_goal[i] < 0:
			target_marble.visible = false
		else:
			target_marble.set_color(puzzle_goal[i])


func _on_marble_entering_goal(body) -> void:
	if body is Marble:
		basket_marbles_color.append(body.get_color_id())

func end_puzzle():
	var condition = verify()
	if condition:
		$"/root/SettingsData".unlock_puzzles[puzzle_number] = true
		$"/root/SettingsData".save_settings()
		if puzzle_number == last_puzzle:
			$"../PopupSolved/VBoxContainer/HBoxContainer/NextBtn".visible = false
		$"../PopupSolved".popup_centered()
	else:
		$"../PopupFail".popup_centered()

			
func verify():
	if open_gate_condition:
		var gates_state = $Board.read_gates()
		var solver = solver_src.new()
		var launch_line = solver.find_open_path(gates_state) 
		if launch_line >= 0:
			return false
		
	if len(basket_marbles_color) < 1:
		return false
	elif puzzle_goal[1] < 0 and puzzle_goal[2] < 0 \
			and puzzle_goal[0] in basket_marbles_color :
		return true
	elif len(basket_marbles_color) == 2 \
			and puzzle_goal[1] >= 0 \
			and puzzle_goal[2] < 0 \
			and puzzle_goal[0] == basket_marbles_color[0] \
			and puzzle_goal[1] == basket_marbles_color[1] :
		return true
	elif puzzle_goal[1] >= 0 and puzzle_goal[2] >= 0 \
			and puzzle_goal[0] != puzzle_goal[1] \
			and puzzle_goal[0] != puzzle_goal[2] \
			and puzzle_goal[1] != puzzle_goal[2] \
			and puzzle_goal[0] in basket_marbles_color \
			and puzzle_goal[1] in basket_marbles_color \
			and puzzle_goal[2] in basket_marbles_color :
		return true
	
	

# Called every frame. 'delta' is the elapsed time since the previous frame.
func _process(_delta: float) -> void:
	pass

func run_puzzle():
	if AudioServer.is_bus_mute(3):
		AudioServer.set_bus_mute(3, false)
	turn += 1
	turn_running = true
	var launch_script = preload('res://src/launch_zone.gd')
	var marble_scn = preload('res://src/marble.tscn').instantiate()
	marble_scn.set_color_by_name($Sources.color)
	 # name.to_int() transform the numbers in the name to int
	var x = launch_script.x0 + launch_script.x_increment * launch_input_choice.to_int()
	marble_scn.position = Vector2(x + $Board.position.x, launch_script.y0 + $Board.position.y)
	marble_scn.name = "MarbleTurn_" + str(turn)
	add_child(marble_scn)
	$BotonConfirm.disabled = true
	starting_time = Time.get_ticks_msec()
	
func _physics_process(_delta: float) -> void:
	if turn_running:
		var any_linear_movement = false
		
		var lenght2_max = 0.0
		# Loop over the marbles
		for marble in get_children():
			if marble is Marble and not marble.felt_to_bottom:
				var vv = marble.get_linear_velocity()
				var l2 = vv.length_squared()
				any_linear_movement = l2 > threshold or any_linear_movement
				lenght2_max = max(lenght2_max, l2)
				
				
		if not any_linear_movement:
			counting_stop_turn += 1
		else:
			counting_stop_turn = 0
				
		if counting_stop_turn > 4:
			end_turn.emit()

func show_menu():
	$"/root/SettingsData".save_settings()
	$HUD.show()
	get_tree().paused = true
	$HUD/HBoxContainer/VBoxContainer/BotonResume.disabled = false
	$"/root/BackgroundMusic".pause_puzzle_music()
	$"/root/BackgroundMusic".play_menu_music()
