extends Node2D

signal gotopuzzle(puzzle_name)
var solver_src = preload('res://src/solo_mode/analytic_solver.gd')
var marble_gd = preload('res://src/marble.gd')
const marble_animation = preload("res://src/marble_animation.tscn")
signal end_turn
var puzzle_ended = false
var last_puzzle = 12
@export var puzzle_number:int = -1
@export var setup_gates:Array = [
	  [0,0,0],
	 [0,0,0,0],
	  [0,0,0],
	 [0,0,0,0]
]
@export var setup_marbles:Array = [
		   [null, null, null],
		[null, null, null, null],
		   [null, null, null],
		[null, null, null, null]
]

@export var puzzle_goal : Array = [-1, -1, -1]
@export var puzzle_opponent : Array = [-1, -1, -1]

const VictoryScore:int = 3

var p1turn : bool
var p2turn : bool
var p1score:int = 0
var p2score:int = 0
var starting_time :int
var turn = 0
var turn_running = false
var launch_input_choice = null
var counting_stop_turn = 0
var basket_marbles_color = []

const epsilon = 750. # TODO: Rename epsilon by threshold
const ang_epsilon = 0.22 # TODO : and refactor to avoid repetition

# Called when the node enters the scene tree for the first time.
func _ready() -> void:
	TranslationServer.set_locale($"/root/SettingsData".lang)
	$BotonConfirm.hide()
	$"/root/BackgroundMusic".pause_menu_music()
	$"/root/BackgroundMusic".pause_puzzle_music()
	$"/root/BackgroundMusic".play_puzzle_music()
	var validation = find_child("Validation") 
	validation.body_entered.connect(_on_marble_entering_goal)
	end_turn.connect(end_puzzle)
	%Title.text = tr("PZZLE_N") + " " + str(puzzle_number)
	%Condition.text = "[center]%s[/center]" % tr("MateIn2")
	$Board.setup_puzzle(setup_gates, setup_marbles, $Board.position.x, $Board.position.y)
	$FancyScoreBoard1.target_marbles = puzzle_goal
	$FancyScoreBoard2.target_marbles = puzzle_opponent
	p1turn = true
	p2turn = false
	puzzle_ended = false
	if puzzle_goal[0] < 0:
		p1score = 1
	if puzzle_goal[1] < 0:
		p1score = 2
	if puzzle_opponent[0] < 0:
		p2score = 1
	if puzzle_opponent[1] < 0:
		p2score = 2
	$FancyScoreBoard1.setup_initial_score(p1score)
	$FancyScoreBoard2.setup_initial_score(p2score)

func _on_marble_entering_goal(body) -> void:
	if body is Marble:
		basket_marbles_color.append(body.get_color_id())

func end_puzzle():
	if not puzzle_ended:
		var puzzle_really_ended = turn > 2 or p1score >= VictoryScore or p2score >= VictoryScore
		if not puzzle_really_ended:
			return
	
	puzzle_ended = true
	
	var right_solution = turn < 4 and p1score >= VictoryScore and p2score < VictoryScore
	if right_solution:
		$"/root/SettingsData".unlock_puzzles[puzzle_number] = true
		$"/root/SettingsData".save_settings()
		if puzzle_number == last_puzzle:
			$"../PopupSolved/VBoxContainer/HBoxContainer/NextBtn".visible = false
		$"../PopupSolved".popup_centered()
	else:
		$"../PopupFail".popup_centered()

		
# Called every frame. 'delta' is the elapsed time since the previous frame.
func _process(_delta: float) -> void:
	if puzzle_ended:
		end_puzzle()

func start_turn():
	if AudioServer.is_bus_mute(3):
		AudioServer.set_bus_mute(3, false)
	turn += 1
	turn_running = true
	counting_stop_turn = 0
	var launch_script = preload('res://src/launch_zone.gd')
	var marble_scn = preload('res://src/marble.tscn').instantiate()
	marble_scn.set_color_by_name($Sources.color)
	 # name.to_int() transform the numbers in the name to int
	var x = launch_script.x0 + launch_script.x_increment * launch_input_choice.to_int()
	marble_scn.position = Vector2(x + $Board.position.x, launch_script.y0 + $Board.position.y)
	marble_scn.name = "MarbleTurn_" + str(turn)
	add_child(marble_scn)
	$BotonConfirm.disabled = true
	hide_input(true)
	starting_time = Time.get_ticks_msec()
	
func _physics_process(_delta: float) -> void:
	if turn_running:
		var any_linear_movement = false
		var any_angular_movement = false
		
		var lenght2_max = 0.0
		var ang_max = 0.0
		
		# Loop over the marbles
		for node in get_children():
			if node.is_class("RigidBody2D"):
				var vv = node.get_linear_velocity()
				var l2 = vv.length_squared()
				any_linear_movement = l2 > epsilon or any_linear_movement
				lenght2_max = max(lenght2_max, l2)
				
		# Loop over the gates
		for node in $Board.get_children():
			if node.is_class("RigidBody2D"):
				var av = abs(node.get_angular_velocity())
				any_angular_movement = av > ang_epsilon or any_angular_movement
				ang_max = max(ang_max, av)
				
		if not any_angular_movement and not any_linear_movement:
			counting_stop_turn += 1
		else:
			counting_stop_turn = 0
				
		if counting_stop_turn > 4:
			await cambiar_de_turno()
			end_turn.emit()

func cambiar_de_turno():
	turn_running = false
	launch_input_choice = null
	$Sources.color = null
	$BotonConfirm.hide()
	### Calculate the score
	if p1turn:
		var board_results = await $FancyScoreBoard1.check_score(p1score)
		p1score = board_results["points"]
	elif p2turn:
		var board_results = await $FancyScoreBoard2.check_score(p2score)
		p2score =  board_results["points"]
		
	### Prapere new turn
	### -----------------
	## Deselect launch line
	launch_input_choice = null
	for zone in $LaunchSelect.get_children():
		zone.disabled = false
		
	## Deselect color
	$Sources.color = null
	for source in $Sources.get_children():
		source.selected = false
		source.disabled = false
	hide_input(false)
		
	await clean_marbles()
	
	### Change player turn
	p1turn = not p1turn
	p2turn = not p2turn
	
	if p2turn and turn < 3: # AI turn
		$BotonConfirm.show()
		$BotonConfirm.set_disabled(true)
		var AI_output = []
		if puzzle_number == 10:
			AI_output =  _AI_calculation_NRM()
		else:
			AI_output = _AI_puzzle()
		
		launch_input_choice = AI_output[0]
		$Sources.color = AI_output[1]

		start_turn()

func clean_marbles():
	for marble in get_children():
		if is_instance_valid(marble)  and marble is Marble and marble.felt_to_bottom:
			print("»»» Name:", marble.name)
			var initial_position = marble.position - self.position
			var marble_color = marble.get_color()
			remove_child(marble)
			marble.queue_free()
			var magic = marble_animation.instantiate()
			add_child(magic)
			var marble_animation_skin = magic.get_child(0) #Skin
			var marble_animation_effect = magic.get_child(1) #Effect
			marble_animation_effect.position = initial_position
			marble_animation_skin.set_color_by_name(marble_color)
			var animatron = marble_animation_skin.get_child(0)
			var anime = animatron.get_animation("disappearing")
			anime.track_set_key_value(0,0, initial_position)
			animatron.play("disappearing")
			await animatron.animation_finished
	
func hide_input(to_hide):
	if to_hide:
		$Sources.hide()
		$LaunchSelect.hide()
	else: 
		$Sources.show()
		$LaunchSelect.show()

func show_menu():
	$"/root/SettingsData".save_settings()
	$HUD.show()
	get_tree().paused = true
	$HUD/HBoxContainer/VBoxContainer/BotonResume.disabled = false
	$"/root/BackgroundMusic".pause_puzzle_music()
	$"/root/BackgroundMusic".play_menu_music()
	
func _AI_puzzle():
	var solver = solver_src.new()
	var gates_state = $Board.read_gates()
	var marbles_state = $Board.read_marbles()
	var gates_to_verify = range(6)
	var launch_line = -1
	var color_index = -1
	var max_possible_score = p2score
	var nxt_player_max_score = p1score
	var min_hmn_score = VictoryScore
	var max_clean_up = 0
	
	
	var win_break = false
	
	for g in gates_to_verify:
		for c in marble_gd.colors:
			var prediction = solver.tensor_next_position(g, c, gates_state, marbles_state)
			var possible_marbles = prediction["basket_marbles"]
			var new_score = $FancyScoreBoard2.calculate_score(p2score, possible_marbles)
			var next_player_gates = prediction["gates"]
			var next_player_marbles = prediction["marbles"]
			
			if new_score >= VictoryScore:
				max_possible_score = new_score
				min_hmn_score = -1
				launch_line = g
				color_index = c
				win_break = true
				break
				
			nxt_player_max_score = p1score
			for g1 in gates_to_verify:
				for c1 in marble_gd.colors:
					var two_turns_prediction = solver.tensor_next_position(g1, c1, next_player_gates, next_player_marbles)
					var nxt_player_forseen_marbles = two_turns_prediction["basket_marbles"]
					var nxt_player_forseen_score = $FancyScoreBoard1.calculate_score(p1score, nxt_player_forseen_marbles)
					if nxt_player_forseen_score > nxt_player_max_score:
						nxt_player_max_score = nxt_player_forseen_score
					
			print("gate: ", g, " color: ", c)
			print(" new_score = ", new_score)
			print(" nxt_player_max_score = ", nxt_player_max_score)
			print(" max_clean_up = ", len(possible_marbles))
			print("--------------------------------------------")
			if new_score >= max_possible_score \
					and nxt_player_max_score <= min_hmn_score\
					and max_clean_up <= len(possible_marbles):
			#if new_score > max_possible_score:
				min_hmn_score =  min(nxt_player_max_score, VictoryScore)
				max_possible_score = new_score
				max_clean_up = len(possible_marbles)
				launch_line = g
				color_index = c
			elif launch_line <  0 and color_index < 0 and len(possible_marbles) > 0:
				launch_line = g
				color_index = c
				
		if win_break:
			break
	
	if launch_line < 0:
		launch_line = solver.random_first_level_open(gates_state)
	if color_index < 0:
		color_index = $FancyScoreBoard2.target_marbles[p2score]
	var output_location = "LaunchZone" + str(launch_line + 1)
	var output_color = $FancyScoreBoard2.color_names[color_index]
	return [output_location, output_color]
		
func _AI_calculation_NRM():
	var solver = solver_src.new()
	var gates_state = $Board.read_gates()
	var marbles_state = $Board.read_marbles()
	var gates_to_verify = solver.open_lvl0_gates(gates_state)
	var launch_line = -1
	var color_index = -1
	var max_possible_score = p2score
	
	for g in gates_to_verify:
		for c in marble_gd.colors:
			var prediction = solver.next_position(g, c, gates_state, marbles_state)
			var possible_marbles = prediction["basket_marbles"]
			var new_score = $FancyScoreBoard2.calculate_score(p2score, possible_marbles)
			if new_score > max_possible_score:
				max_possible_score = new_score
				launch_line = g
				color_index = c
			elif launch_line <  0 and color_index < 0 and len(possible_marbles) > 0:
				launch_line = g
				color_index = c
				
			
	
	if launch_line < 0:
		launch_line = solver.random_first_level_open(gates_state)
	if color_index < 0:
		color_index = $FancyScoreBoard2.target_marbles[p2score]
	var output_location = "LaunchZone" + str(launch_line + 1)
	var output_color = $FancyScoreBoard2.color_names[color_index]
	return [output_location, output_color]


func _on_boton_reload_pressed() -> void:
	gotopuzzle.emit(puzzle_number)
