extends Node2D

class_name PvP

signal end_turn
signal marble_jam

const marble_animation = preload("res://src/marble_animation.tscn")

var marbles_to_handle = []

var turn = 0
var jam_count = 0
const time_limit_jammed =  10000
const time_long_enough =  5000
var dealing_with_marble_jam = false

const threshold = 760. 

# Declare member variables here. 
var starting_time :int
var held_object = null
var turn_running = false
var launch_input_choice = null

var counting_stop_turn = 0

var p1score:int = 0
var p2score:int = 0
var p1turn:bool = true
var p2turn:bool = false
var p1name:String = "null"
var p2name:String = "null"
const VictoryScore:int = 9


const x_cleanning:float = 375.0 
const y_cleanning_right:float = 225.0 
const y_cleanning:float = 840.0 


func _ready():
	TranslationServer.set_locale($"/root/SettingsData".lang)
	$BotonConfirm.hide()
	$"/root/BackgroundMusic".pause_menu_music()
	$"/root/BackgroundMusic".play_game_music()
	
	## Start the scene PvP
	p1score = 0
	p2score = 0
	p1turn = true
	p2turn = false
	update_players_name()
	cleanup_all()
	
	randomize()
	$Board.setup_gates(true, null) # true -> Random
		
	$Turno.text = tr("TURN_OF") + p1name
	$Puntuacion_Player1.text = p1name + " : " + str(p1score) + "pts"
	$Puntuacion_Player2.text = p2name + " : " + str(p2score) + "pts"
	$FancyScoreBoard1.setup()
	$FancyScoreBoard2.setup()
	$FancyScoreBoard1.queue_redraw()
	$FancyScoreBoard2.queue_redraw()

func start_turn():
	if AudioServer.is_bus_mute(3):
		AudioServer.set_bus_mute(3, false)
	turn += 1
	turn_running = true
	dealing_with_marble_jam = false
	jam_count = 0
	counting_stop_turn = 0
	var launch_script = preload('res://src/launch_zone.gd')
	var marble_scn = preload('res://src/marble.tscn').instantiate()
	marble_scn.set_color_by_name($Sources.color)
	 # name.to_int() transform the numbers in the name to int
	var x = launch_script.x0 + launch_script.x_increment * launch_input_choice.to_int()
	marble_scn.position = Vector2(x, launch_script.y0)
	marble_scn.name = "MarbleTurn_" + str(turn)
	add_child(marble_scn)
	$BotonConfirm.disabled = true
	hide_input(true)
	starting_time = Time.get_ticks_msec()

func hide_input(to_hide):
	if to_hide:
		$Sources.hide()
		$LaunchSelect.hide()
	else: 
		$Sources.show()
		$LaunchSelect.show()

func show_menu():
	$"/root/BackgroundMusic".pause_game_music()
	$"/root/BackgroundMusic".play_menu_music()
	get_tree().paused = true
	$HUD.show()
	$HUD/HBoxContainer/VBoxContainer/BotonResume.disabled = false

func _physics_process(_delta: float) -> void:
	if turn_running:
		var any_linear_movement = false
		
		var lenght2_max = 0.0

		# Loop over the marbles
		for marble in get_children():
			if marble is Marble and not marble.felt_to_bottom:
				var vv = marble.get_linear_velocity()
				var l2 = vv.length_squared()
				any_linear_movement = l2 > threshold or any_linear_movement
				lenght2_max = max(lenght2_max, l2)
				
				
		if not any_linear_movement:
			counting_stop_turn += 1
		else:
			counting_stop_turn = 0
		
		var elapsed_time = Time.get_ticks_msec() - starting_time 
		if counting_stop_turn > 6:
			await cambiar_de_turno()
			end_turn.emit()
		elif dealing_with_marble_jam:
			for marble in get_children():
				if marble is Marble and marble.to_be_shaken:
					var animatron = marble.find_child("AnimationPlayer")
					animatron.play("shake")
					marble.set_linear_velocity(Vector2(0.0,200.0))
					marble.set_angular_velocity(90.0)
					marble.to_be_shaken = false

		elif not dealing_with_marble_jam and elapsed_time > time_limit_jammed:
			dealing_with_marble_jam = true
			await counting_points()
			marble_jam.emit()
			await counting_points()
		elif not dealing_with_marble_jam and elapsed_time > time_long_enough:
			# TODO: New button
			pass

func game_over(winner):
	$VictoryPanel.show()
	# $VictoryPanel/Trumpet_Left/Skeleton2D/Root/Bone2D/Bone2D/AnimationPlayer.play("sound")
	# $VictoryPanel/Trumpet_Right/Skeleton2D/Root/Bone2D/Bone2D/AnimationPlayer.play("sound")
	$VictoryPanel/Tadaaa.play()
	if $"/root/SettingsData".lang == "es":
		$VictoryPanel/Label.text = "¡" + winner + " " + tr("VICTORY") + "!"
	else :
		$VictoryPanel/Label.text = winner + " " + tr("VICTORY") + "!"
	$VictoryButton.show()

func change_settings():
	update_players_name()
	update_score_text()
	update_turn_text()

func update_score_text():
	$Puntuacion_Player1.text = p1name + " : " + str(p1score) + "pts"
	$Puntuacion_Player2.text = p2name + " : " + str(p2score) + "pts"


func update_turn_text():
	if p2turn:
		$Turno.text = tr("TURN_OF") + p2name
	elif p1turn:
		$Turno.text = tr("TURN_OF") + p1name
	

func update_players_name():
	p1name = $"/root/SettingsData".player1_name
	p2name = $"/root/SettingsData".player2_name
	if p1name == "" :
		p1name = tr("PLAYER") + " 1"
	if p2name == "" :
		p2name = tr("PLAYER") + " 2"

func verify_score():
	### Calculate the score
	if p1turn:
		var board_results = await $FancyScoreBoard1.check_score(p1score)
		p1score = board_results["points"]
	elif p2turn:
		var board_results = await $FancyScoreBoard2.check_score(p2score)
		p2score =  board_results["points"]

func counting_points():
	await verify_score()

	if p1score >= VictoryScore:
		game_over(p1name)
		return
	elif p2score >= VictoryScore:
		game_over(p2name)
		return
	
	update_score_text()

func cambiar_de_turno():
	turn_running = false
	launch_input_choice = null
	$Sources.color = null
	$BotonConfirm.hide()
	await counting_points()
	
	### Prapere new turn
	### -----------------
	## Deselect launch line
	launch_input_choice = null
	for zone in $LaunchSelect.get_children():
		zone.disabled = false
		
	## Deselect color
	$Sources.color = null
	for source in $Sources.get_children():
		source.selected = false
		source.disabled = false
	hide_input(false)
		
	await clean_marbles()
	
	### Change player turn
	p1turn = not p1turn
	p2turn = not p2turn
	update_turn_text()

func clean_marbles():
	for marble in get_children():
		if is_instance_valid(marble) and marble is Marble and marble.felt_to_bottom:
			var initial_position = marble.position - self.position
			var marble_color = marble.get_color()
			remove_child(marble)
			marble.queue_free()
			var magic = marble_animation.instantiate()
			add_child(magic)
			var marble_animation_skin = magic.get_child(0) #Skin
			var marble_animation_effect = magic.get_child(1) #Effect
			marble_animation_effect.position = initial_position
			marble_animation_skin.set_color_by_name(marble_color)
			var animatron = marble_animation_skin.get_child(0)
			var anime = animatron.get_animation("disappearing")
			anime.track_set_key_value(0,0, initial_position)
			animatron.play("disappearing")
			await animatron.animation_finished

func cleanup_all():
	# TODO: I think this code does nothing. Clean up
	for node in get_children():
		if node.is_class("Gate") or \
		  node.is_class("Marble"):
			remove_child(node)
			node.queue_free()


func _on_marble_jam() -> void:
	if jam_count == 0:
		marbles_to_handle = []
		for node in $Board/BackBoard.get_children():
			if node.name.begins_with("GateSection"):
				if node.marbles.size() > 1:
					for marble in node.marbles:
						marble.to_be_shaken = true
						marbles_to_handle.append(marble)
						
	if marbles_to_handle.size() == 0:
		return
		
	jam_count += 1
	#for marble in marbles_to_handle:
	#	var handle_marble_scn = preload('res://src/handle_the_marble.tscn').instantiate()
	#	add_child(handle_marble_scn)
	#	handle_marble_scn.marble_info(marble)
	#	marbles_to_handle.erase(marble)


func _on_victory_button_pressed() -> void:
	get_tree().change_scene_to_file("res://src/thanks.tscn")
