extends Node

const SAVE_GAME_SETTINGS = "user://settings.json"
var lang = "en"
var player1_name = ""
var player2_name = ""
var difficulty = 0
var music_volume:int = 70
var sound_volume:int = 70
var unlock_puzzles = [
	true, false, false, 
	false, false, false,
	false, false, false,
	false, false, false,
	false, false, false,
	false, false, false
	]

var file = null

func save_settings():
	var data = {
		"language" : lang,
		"player1" : player1_name,
		"player2" : player2_name,
		"unlock_puzzles" : unlock_puzzles,
		"music_volume" : music_volume,
		"sound_volume" : sound_volume,
	}
	var data_string = str(JSON.stringify(data))
	file = FileAccess.open(SAVE_GAME_SETTINGS, FileAccess.WRITE)
	file.store_string(data_string)

func load_settings()-> bool:
	if not FileAccess.file_exists(SAVE_GAME_SETTINGS):
		lang = "en"
		music_volume = 70
		sound_volume = 70
		player1_name = ""
		player2_name = ""
		unlock_puzzles = [
			true, false, false, 
			false, false, false,
			false, false, false,
			false, false, false,
			false, false, false,
			false, false, false
			]
		update_volume(music_volume,1)
		update_volume(sound_volume,2)
		return false
	else:
		file = FileAccess.open(SAVE_GAME_SETTINGS, FileAccess.READ)
		var data_string = file.get_as_text()
		
		var json = JSON.new()
		var error = json.parse(data_string)
		
		if error == OK:
			var data_received = json.data
			lang = data_received["language"]
			player1_name = data_received["player1"]
			player2_name = data_received["player2"]
			
			if "music_volume" in data_received:
				music_volume = data_received["music_volume"]
			else:
				music_volume = 70
			if "sound_volume" in data_received:
				sound_volume = data_received["sound_volume"]
			else:
				sound_volume = 70
				
			if "unlock_puzzles" in data_received:
				unlock_puzzles = data_received["unlock_puzzles"]
			else:
				unlock_puzzles = [
					true, false, false, 
					false, false, false,
					false, false, false,
					false, false, false,
					false, false, false,
					false, false, false
					]
		
		TranslationServer.set_locale(lang)
		update_volume(music_volume,1)
		update_volume(sound_volume,2)
		return true

func update_volume(local_volume: int, local_bus_index:int) -> void:
	if local_volume == 0:
		# Mute
		AudioServer.set_bus_mute(local_bus_index, true)
	else:
		# Unmute first
		if AudioServer.is_bus_mute(local_bus_index):
			AudioServer.set_bus_mute(local_bus_index, false)
		var new_db = convert_to_db(local_volume)
		AudioServer.set_bus_volume_db(local_bus_index, new_db)

func convert_to_db(percentage:int) -> float:
	var db_scale = 40.0
	var divisor = 70.0
	var db = db_scale * log(percentage/divisor) / log(10.0)
	return db
