extends Node

var marble = preload('res://src/marble.gd')


func next_position(
		launch_zone: int,
		marble_color: int,
		setup_gates: Array, 
		setup_marbles: Array
) -> Dictionary:
	var gates_state = setup_gates.duplicate(true)  # Deep copy
	var marbles_state = setup_marbles.duplicate(true)  # Deep copy
	var output_marbles = []
	
	#LEVEL0
	var input_marble = [null, null, null, null, null, null, null, null]
	input_marble[launch_zone+1]=marble_color # +1 because input marble is shape of 8 not 6
	var lvl0_prediction = next_position_level0(
		input_marble, 
		gates_state, 
		marbles_state,
		0 # level
		)
	gates_state = lvl0_prediction["gates"]
	marbles_state = lvl0_prediction["marbles"]
	var marbles2lvl1 = lvl0_prediction["next_lvl"]
	
	#LEVEL1
	var lvl1_prediction = next_position_level1(
		marbles2lvl1, 
		gates_state, 
		marbles_state,
		1 # level
		)
	gates_state = lvl1_prediction["gates"]
	marbles_state = lvl1_prediction["marbles"]
	var marbles2lvl2 = lvl1_prediction["next_lvl"]
	
	#LEVEL2
	var lvl2_prediction = next_position_level0(
		marbles2lvl2, 
		gates_state, 
		marbles_state,
		2 # level
		)
	gates_state = lvl2_prediction["gates"]
	marbles_state = lvl2_prediction["marbles"]
	var marbles2lvl3 = lvl2_prediction["next_lvl"]

	#LEVEL3
	var lvl3_prediction = next_position_level1(
		marbles2lvl3, 
		gates_state, 
		marbles_state,
		3 # level
		)
	gates_state = lvl3_prediction["gates"]
	marbles_state = lvl3_prediction["marbles"]
	var marbles2basket = lvl3_prediction["next_lvl"]
	for i in marbles2basket.size():
		if marbles2basket[i] != null:
			output_marbles.append(marbles2basket[i])
	return {
		"gates": gates_state,
		"marbles": marbles_state,
		"basket_marbles" : output_marbles
		}


func next_position_level0(
		input_marbles: Array,
		gates_state: Array, 
		marbles_state: Array,
		level: int
) -> Dictionary:
	var prev_gates_lvl0 = gates_state[level]
	var prev_marbles_lvl0 = marbles_state[level]
	var marbles_to_next_lvl = [null, null, null, null, null, null, null, null]
	marbles_to_next_lvl[0] = input_marbles[0]
	marbles_to_next_lvl[7] = input_marbles[7]
	for i in [0,1,2]:
		if input_marbles[2*i+1]  != null and input_marbles[2*i+2]  != null:
			marbles_to_next_lvl[2*i+1] = input_marbles[2*i+1]
			marbles_to_next_lvl[2*i+2] = input_marbles[2*i+2]
			marbles_state[level][i] = null
		elif input_marbles[2*i+1]  != null:
			if prev_marbles_lvl0[i] != null:
				if prev_gates_lvl0[i] == -1:
					marbles_to_next_lvl[2*i+2] = input_marbles[2*i+1]
					marbles_to_next_lvl[2*i+1] = prev_marbles_lvl0[i]
					marbles_state[level][i] = null
				elif prev_gates_lvl0[i] == 1:
					marbles_to_next_lvl[2*i+1] = input_marbles[2*i+1]
					marbles_to_next_lvl[2*i+2] = prev_marbles_lvl0[i]
					marbles_state[level][i] = null
				
			elif prev_gates_lvl0[i] == 0 and prev_marbles_lvl0[i] == null:
				gates_state[level][i] = -1
				marbles_state[level][i] = input_marbles[2*i+1]
			elif prev_gates_lvl0[i] == -1 and prev_marbles_lvl0[i] == null:
				gates_state[level][i] = -1
				marbles_state[level][i] = input_marbles[2*i+1]
			elif prev_gates_lvl0[i] == 1 and prev_marbles_lvl0[i] == null:
				gates_state[level][i] = -1
				marbles_state[level][i] = null
				marbles_to_next_lvl[2*i+1] = input_marbles[2*i+1]
		elif input_marbles[2*i+2]  != null:
			if prev_marbles_lvl0[i] != null:
				if prev_gates_lvl0[i] == -1:
					marbles_to_next_lvl[2*i+2] = input_marbles[2*i+2]
					marbles_to_next_lvl[2*i+1] = prev_marbles_lvl0[i]
					marbles_state[level][i] = null
				elif prev_gates_lvl0[i] == 1:
					marbles_to_next_lvl[2*i+1] = input_marbles[2*i+2]
					marbles_to_next_lvl[2*i+2] = prev_marbles_lvl0[i]
					marbles_state[level][i] = null
			elif prev_gates_lvl0[i] == 0 and prev_marbles_lvl0[i] == null:
				gates_state[level][i] = 1
				marbles_state[level][i] = input_marbles[2*i+2]
			elif prev_gates_lvl0[i] == 1 and prev_marbles_lvl0[i] == null:
				gates_state[level][i] = 1
				marbles_state[level][i] = input_marbles[2*i+2]
			elif prev_gates_lvl0[i] == -1 and prev_marbles_lvl0[i] == null:
				gates_state[level][i] = 1
				marbles_state[level][i] = null
				marbles_to_next_lvl[2*i+2] = input_marbles[2*i+2]

	return {
		"gates": gates_state,
		"marbles": marbles_state,
		"next_lvl": marbles_to_next_lvl
		}

func next_position_level1(
		input_marbles: Array,
		gates_state: Array, 
		marbles_state: Array,
		level: int
) -> Dictionary:
	var prev_gates_lvl1 = gates_state[level]
	var prev_marbles_lvl1 = marbles_state[level]
	var marbles_to_next_lvl = [null, null, null, null, null, null, null, null]
	for i in [0,1,2,3]:
		if input_marbles[2*i] != null and input_marbles[2*i+1] != null:
			marbles_to_next_lvl[2*i] = input_marbles[2*i]
			marbles_to_next_lvl[2*i+1] = input_marbles[2*i+1]
			marbles_state[level][i] = null
		elif input_marbles[2*i] != null:
			if prev_marbles_lvl1[i] != null:
				if prev_gates_lvl1[i] == -1:
					marbles_to_next_lvl[2*i] = prev_marbles_lvl1[i]
					marbles_to_next_lvl[2*i+1] = input_marbles[2*i]
					marbles_state[level][i] = null
				elif prev_gates_lvl1[i] == 1:
					marbles_to_next_lvl[2*i] = input_marbles[2*i]
					marbles_to_next_lvl[2*i+1] = prev_marbles_lvl1[i]
					marbles_state[level][i] = null
			elif prev_gates_lvl1[i] == -1: # and prev_marbles_lvl1[i] == null:
				gates_state[level][i] = -1
				marbles_state[level][i] = input_marbles[2*i]
			elif prev_gates_lvl1[i] == 1: # and prev_marbles_lvl1[i] == null:
				gates_state[level][i] = -1
				marbles_to_next_lvl[2*i] = input_marbles[2*i]
		elif input_marbles[2*i+1] != null:
			if prev_marbles_lvl1[i] != null:
				if prev_gates_lvl1[i] == -1:
					marbles_to_next_lvl[2*i] = prev_marbles_lvl1[i]
					marbles_to_next_lvl[2*i+1] = input_marbles[2*i+1]
					marbles_state[level][i] = null
				elif prev_gates_lvl1[i] == 1:
					marbles_to_next_lvl[2*i] = input_marbles[2*i+1]
					marbles_to_next_lvl[2*i+1] = prev_marbles_lvl1[i]
					marbles_state[level][i] = null
			elif prev_gates_lvl1[i] == 1 and prev_marbles_lvl1[i] == null:
				gates_state[level][i] = 1
				marbles_state[level][i] = input_marbles[2*i+1]
			elif prev_gates_lvl1[i] == -1 and prev_marbles_lvl1[i] == null:
				gates_state[level][i] = 1
				marbles_to_next_lvl[2*i+1] = input_marbles[2*i+1]
		
	return {
		"gates": gates_state,
		"marbles": marbles_state,
		"next_lvl": marbles_to_next_lvl
		}


func tensor_next_position(
		launch_zone: int,
		marble_color: int,
		setup_gates: Array, 
		setup_marbles: Array
) -> Dictionary:
	var gates_state = setup_gates.duplicate(true)  # Deep copy
	var marbles_state = setup_marbles.duplicate(true)  # Deep copy
	var output_marbles = []
	
	#LEVEL0
	var input_marble = [[], [], [], [], [], [], [], []]
	input_marble[launch_zone+1]= [marble_color] # +1 because input marble is shape of 8 not 6
	var lvl0_prediction = tensor_next_position_level0(
		input_marble, 
		gates_state, 
		marbles_state,
		0 # level
		)
	gates_state = lvl0_prediction["gates"]
	marbles_state = lvl0_prediction["marbles"]
	var marbles2lvl1 = lvl0_prediction["next_lvl"]
	
	#LEVEL1
	var lvl1_prediction = tensor_next_position_level1(
		marbles2lvl1, 
		gates_state, 
		marbles_state,
		1 # level
		)
	gates_state = lvl1_prediction["gates"]
	marbles_state = lvl1_prediction["marbles"]
	var marbles2lvl2 = lvl1_prediction["next_lvl"]
	
	#LEVEL2
	var lvl2_prediction = tensor_next_position_level0(
		marbles2lvl2, 
		gates_state, 
		marbles_state,
		2 # level
		)
	gates_state = lvl2_prediction["gates"]
	marbles_state = lvl2_prediction["marbles"]
	var marbles2lvl3 = lvl2_prediction["next_lvl"]

	#LEVEL3
	var lvl3_prediction = tensor_next_position_level1(
		marbles2lvl3, 
		gates_state, 
		marbles_state,
		3 # level
		)
	gates_state = lvl3_prediction["gates"]
	marbles_state = lvl3_prediction["marbles"]
	
	for possible_marbles in lvl3_prediction["next_lvl"]:
		output_marbles.append_array(possible_marbles)
	return {
		"gates": gates_state,
		"marbles": marbles_state,
		"basket_marbles" : output_marbles
		}


func tensor_next_position_level0(
		input_marbles: Array,
		gates_state: Array, 
		marbles_state: Array,
		level: int
) -> Dictionary:
	var prev_gates_lvl0 = gates_state[level]
	var prev_marbles_lvl0 = marbles_state[level]
	var marbles_to_next_lvl = [[], [], [], [], [], [], [], []]
	marbles_to_next_lvl[0] = input_marbles[0]
	marbles_to_next_lvl[7] = input_marbles[7]
	for i in [0,1,2]:
		if len(input_marbles[2*i+1])>0 and len(input_marbles[2*i+2])>0:
			marbles_to_next_lvl[2*i+1].append_array(input_marbles[2*i+1])
			marbles_to_next_lvl[2*i+2].append_array(input_marbles[2*i+2])
			marbles_state[level][i] = null
		elif len(input_marbles[2*i+1] )>0:
			if prev_marbles_lvl0[i] != null:
				if prev_gates_lvl0[i] == -1:
					marbles_to_next_lvl[2*i+2].append_array(input_marbles[2*i+1])
					marbles_to_next_lvl[2*i+1].append(prev_marbles_lvl0[i])
					marbles_state[level][i] = null
				elif prev_gates_lvl0[i] == 1:
					marbles_to_next_lvl[2*i+1].append_array(input_marbles[2*i+1])
					marbles_to_next_lvl[2*i+2].append(prev_marbles_lvl0[i])
					marbles_state[level][i] = null
				
			elif prev_gates_lvl0[i] == 0 and prev_marbles_lvl0[i] == null and len(input_marbles[2*i+1])==1:
				gates_state[level][i] = -1
				marbles_state[level][i] = input_marbles[2*i+1][0]
			elif prev_gates_lvl0[i] == -1 and prev_marbles_lvl0[i] == null and len(input_marbles[2*i+1])==1:
				gates_state[level][i] = -1
				marbles_state[level][i] = input_marbles[2*i+1][0]
			elif prev_gates_lvl0[i] == 1 and prev_marbles_lvl0[i] == null:
				gates_state[level][i] = -1
				marbles_state[level][i] = null
				marbles_to_next_lvl[2*i+1].append_array(input_marbles[2*i+1])
		elif len(input_marbles[2*i+2])>0:
			if prev_marbles_lvl0[i] != null:
				if prev_gates_lvl0[i] == -1:
					marbles_to_next_lvl[2*i+2].append_array(input_marbles[2*i+2])
					marbles_to_next_lvl[2*i+1].append(prev_marbles_lvl0[i])
					marbles_state[level][i] = null
				elif prev_gates_lvl0[i] == 1:
					marbles_to_next_lvl[2*i+1].append_array(input_marbles[2*i+2])
					marbles_to_next_lvl[2*i+2].append(prev_marbles_lvl0[i])
					marbles_state[level][i] = null
			elif prev_gates_lvl0[i] == 0 and prev_marbles_lvl0[i] == null and  len(input_marbles[2*i+2])==1:
				gates_state[level][i] = 1
				marbles_state[level][i] = input_marbles[2*i+2][0]
			elif prev_gates_lvl0[i] == 1 and prev_marbles_lvl0[i] == null and len(input_marbles[2*i+2])==1:
				gates_state[level][i] = 1
				marbles_state[level][i] = input_marbles[2*i+2][0] 
			elif prev_gates_lvl0[i] == -1 and prev_marbles_lvl0[i] == null:
				gates_state[level][i] = 1
				marbles_state[level][i] = null
				marbles_to_next_lvl[2*i+2].append_array(input_marbles[2*i+2])

	return {
		"gates": gates_state,
		"marbles": marbles_state,
		"next_lvl": marbles_to_next_lvl
		}


func tensor_next_position_level1(
		input_marbles: Array,
		gates_state: Array, 
		marbles_state: Array,
		level: int
) -> Dictionary:
	var prev_gates_lvl1 = gates_state[level]
	var prev_marbles_lvl1 = marbles_state[level]
	var marbles_to_next_lvl = [[], [], [], [], [], [], [], []]
	for i in [0,1,2,3]:
		if len(input_marbles[2*i])>0 and len(input_marbles[2*i+1])>0:
			marbles_to_next_lvl[2*i].append_array(input_marbles[2*i])
			marbles_to_next_lvl[2*i+1].append_array(input_marbles[2*i+1])
			marbles_state[level][i] = null
		elif len(input_marbles[2*i])>0:
			if prev_marbles_lvl1[i] != null:
				if prev_gates_lvl1[i] == -1:
					marbles_to_next_lvl[2*i].append(prev_marbles_lvl1[i])
					marbles_to_next_lvl[2*i+1].append_array(input_marbles[2*i])
					marbles_state[level][i] = null
				elif prev_gates_lvl1[i] == 1:
					marbles_to_next_lvl[2*i].append_array(input_marbles[2*i])
					marbles_to_next_lvl[2*i+1].append(prev_marbles_lvl1[i])
					marbles_state[level][i] = null
			elif prev_gates_lvl1[i] == -1 and len(input_marbles[2*i]): # and prev_marbles_lvl1[i] == null:
				gates_state[level][i] = -1
				marbles_state[level][i] = input_marbles[2*i][0]
			elif prev_gates_lvl1[i] == -1: # and prev_marbles_lvl1[i] == null:
				gates_state[level][i] = -1
				marbles_to_next_lvl[2*i].append(input_marbles[2*i][0])
				for index in range(1,len(input_marbles[2*i+1])):
					marbles_to_next_lvl[2*i+1].append(input_marbles[2*i][index])
			elif prev_gates_lvl1[i] == 1: # and prev_marbles_lvl1[i] == null:
				gates_state[level][i] = -1
				marbles_to_next_lvl[2*i].append_array(input_marbles[2*i])
		elif len(input_marbles[2*i+1])>0:
			if prev_marbles_lvl1[i] != null:
				if prev_gates_lvl1[i] == -1:
					marbles_to_next_lvl[2*i].append(prev_marbles_lvl1[i])
					marbles_to_next_lvl[2*i+1].append_array(input_marbles[2*i+1])
					marbles_state[level][i] = null
				elif prev_gates_lvl1[i] == 1:
					marbles_to_next_lvl[2*i].append_array(input_marbles[2*i+1])
					marbles_to_next_lvl[2*i+1].append(prev_marbles_lvl1[i])
					marbles_state[level][i] = null
			elif prev_gates_lvl1[i] == 1 and prev_marbles_lvl1[i] == null and len(input_marbles[2*i+1])==1:
				gates_state[level][i] = 1
				marbles_state[level][i] = input_marbles[2*i+1][0]
			elif prev_gates_lvl1[i] == 1 and prev_marbles_lvl1[i] == null:
				gates_state[level][i] = 1
				marbles_to_next_lvl[2*i+1].append(input_marbles[2*i+1][0])
				for index in range(1,len(input_marbles[2*i+1])):
					marbles_to_next_lvl[2*i].append(input_marbles[2*i+1][index])
			elif prev_gates_lvl1[i] == -1 and prev_marbles_lvl1[i] == null:
				gates_state[level][i] = 1
				marbles_to_next_lvl[2*i+1].append_array(input_marbles[2*i+1])
		
	return {
		"gates": gates_state,
		"marbles": marbles_state,
		"next_lvl": marbles_to_next_lvl
		}


func find_open_path(setup_gates: Array) -> int:
	# It sould find first open line, starting from left
	for i in [0,1,2]:
		if setup_gates[0][i] == 1:
			if setup_gates[1][i] == -1 and setup_gates[2][i] == 1 and setup_gates[3][i] == -1:
				return 2*i
		elif setup_gates[0][i] == -1:
			if setup_gates[1][i+1] == 1 and setup_gates[2][i] == -1 and setup_gates[3][i+1] == 1:
				return 2*i+1
	return -1
	
func find_semi_open_path(setup_gates: Array, marbles:Array) -> int:
	# It sould find first semi open line with a marble in the last level, starting from left
	for i in [0,1,2]:
		if setup_gates[0][i] == 1:
			if setup_gates[1][i] == -1 and setup_gates[2][i] == 1 and marbles[3][i] != null:
				return 2*i
		elif setup_gates[0][i] == -1:
			if setup_gates[1][i+1] == 1 and setup_gates[2][i] == -1 and marbles[3][i+1] != null:
				return 2*i+1
	return -1
	
func random_first_level_open(setup_gates: Array)  -> int:
	var i = randi() % 3
	if setup_gates[0][i] == 1:
		return 2*i
	else:
		return 2*i+1
		
func open_lvl0_gates(setup_gates: Array) -> Array:
	var result = [-1,-1,-1]
	for i in setup_gates[0].size():
		result[i] = 2*i + 1 - (setup_gates[0][i]+1)/2
	return result
	
