extends Node2D

signal end_turn
signal marble_jam

const marble_animation = preload("res://src/marble_animation.tscn")

var marbles_to_handle = []

var turn = 0
var jam_count = 0
const time_limit_jammed =  15000
var dealing_with_marble_jam = false

const epsilon = 750. # Rename epsilon by threshold
const ang_epsilon = 0.22

# Declare member variables here. 
var starting_time :int
var held_object = null
var turn_running = false
var launch_input_choice = null

var counting_stop_turn = 0

var p1score:int = 0
var p2score:int = 0
var p1turn:bool = true
var p2turn:bool = false
var p1name:String = "null"
var p2name:String = "nullos"
const VictoryScore:int = 9


const x_cleanning:float = 375.0 
const y_cleanning_right:float = 225.0 
const y_cleanning:float = 840.0 

var step21_complete = false
var step22_complete = false
var step23_complete = false
var step24_complete = false

var setup_to_load = [
	  [1,1,1],
	 [1,-1,1,1],
	  [1,1,1],
	 [-1,1,-1,1]
]

func _ready():
	end_turn.connect(from_step23_to_step24)
	$"/root/BackgroundMusic".pause_menu_music()
	TranslationServer.set_locale($"/root/SettingsData".lang)
	update_lang()
	$BotonConfirm.hide()
	
	## Start the scene PvP
	p1score = 0
	p2score = 0
	p1turn = true
	p2turn = false
	update_players_name()
	cleanup_all()
	
	randomize()
	#$Board.setup_gates(true, null) # true -> Random
		
	$Turno.text = tr("TUTO") + " 4/5"
	$Puntuacion_Player1.text = p1name + " : " + str(p1score) + "pts"
	$Puntuacion_Player2.text = p2name + " : " + str(p2score) + "pts"
	$FancyScoreBoard1.setup()
	$FancyScoreBoard2.setup()
	$FancyScoreBoard1.queue_redraw()
	$FancyScoreBoard2.queue_redraw()
	start_turn()

func start_turn():
	var lane_number = 2
	$Board.setup_gates(false, setup_to_load)
	
	turn += 1
	turn_running = true
	dealing_with_marble_jam = false
	jam_count = 0
	counting_stop_turn = 0
	var launch_script = preload('res://src/launch_zone.gd')
	var marble_scn = preload('res://src/marble.tscn').instantiate()
	marble_scn.set_color_by_name("red") # Index 1
	var color_to_impose = marble_scn.get_color_id() - 1
	
	$FancyScoreBoard1.target_marbles[0] = color_to_impose
	$FancyScoreBoard1.queue_redraw()

	 # name.to_int() transform the numbers in the name to int
	var x = launch_script.x0 + launch_script.x_increment * (lane_number + 1)
	marble_scn.position = Vector2(x, launch_script.y0)
	marble_scn.name = "MarbleTurn_" + str(turn)
	add_child(marble_scn)
	$BotonConfirm.disabled = true
	hide_input(true)
	starting_time = Time.get_ticks_msec()

func hide_input(to_hide):
	if to_hide:
		$Sources.hide()
		$LaunchSelect.hide()
	else: 
		$Sources.show()
		$LaunchSelect.show()

func show_menu():
	$"/root/BackgroundMusic".pause_puzzle_music()
	$"/root/BackgroundMusic".play_menu_music()
	$HUD.show()
	get_tree().paused = true
	$HUD/HBoxContainer/VBoxContainer/BotonResume.disabled = false

func _physics_process(_delta: float) -> void:
	if AudioServer.is_bus_mute(3) and Time.get_ticks_msec() - starting_time > 300:
		AudioServer.set_bus_mute(3, false)
	if turn_running:
		var any_linear_movement = false
		var any_angular_movement = false
		
		var lenght2_max = 0.0
		var ang_max = 0.0
		
		# Loop over the marbles
		for node in get_children():
			if node.is_class("RigidBody2D"):
				var vv = node.get_linear_velocity()
				var l2 = vv.length_squared()
				any_linear_movement = l2 > epsilon or any_linear_movement
				lenght2_max = max(lenght2_max, l2)
				
		# Loop over the gates
		for node in $Board.get_children():
			if node.is_class("RigidBody2D"):
				var av = abs(node.get_angular_velocity())
				any_angular_movement = av > ang_epsilon or any_angular_movement
				ang_max = max(ang_max, av)
				
		if not any_angular_movement and not any_linear_movement:
			counting_stop_turn += 1
		else:
			counting_stop_turn = 0
		
		
		if counting_stop_turn > 4:
			await cambiar_de_turno()
			end_turn.emit()
		elif not dealing_with_marble_jam and Time.get_ticks_msec() - starting_time > time_limit_jammed:
			dealing_with_marble_jam = true
			marble_jam.emit()

func game_over(winner):
	$VictoryPanel.show()
	# $VictoryPanel/Trumpet_Left/Skeleton2D/Root/Bone2D/Bone2D/AnimationPlayer.play("sound")
	# $VictoryPanel/Trumpet_Right/Skeleton2D/Root/Bone2D/Bone2D/AnimationPlayer.play("sound")
	$VictoryPanel/Tadaaa.play()
	if $"/root/SettingsData".lang == "es":
		$VictoryPanel/Label.text = "¡" + winner + " " + tr("VICTORY") + "!"
	else :
		$VictoryPanel/Label.text = winner + " " + tr("VICTORY") + "!"
	$VictoryButton.show()

func change_settings():
	update_players_name()
	update_score_text()

func update_score_text():
	$Puntuacion_Player1.text = p1name + " : " + str(p1score) + "pts"
	$Puntuacion_Player2.text = p2name + " : " + str(p2score) + "pts"


func update_players_name():
	p1name = tr("PLAYER") + " 1"
	p2name = tr("PLAYER") + " 2"


func cambiar_de_turno():
	turn_running = false
	launch_input_choice = null
	$Sources.color = null
	$BotonConfirm.hide()
	### Calculate the score
	if p1turn:
		var board_results = await $FancyScoreBoard1.check_score(p1score)
		p1score = board_results["points"]
	elif p2turn:
		var board_results = await $FancyScoreBoard2.check_score(p2score)
		p2score =  board_results["points"] 

	if p1score >= VictoryScore:
		game_over(p1name)
		return
	elif p2score >= VictoryScore:
		game_over(p2name)
		return
	
	update_score_text()
	
	### Prapere new turn
	### -----------------
	## Deselect launch line
	launch_input_choice = null
	for zone in $LaunchSelect.get_children():
		zone.disabled = false
		
	## Deselect color
	$Sources.color = null
	for source in $Sources.get_children():
		source.selected = false
		source.disabled = false
	hide_input(false)
		
	await clean_marbles()
	
	### Change player turn
	p1turn = not p1turn
	p2turn = not p2turn


func clean_marbles():
	for marble in get_children():
		if is_instance_valid(marble) and marble is Marble and marble.felt_to_bottom:
			var initial_position = marble.position - self.position
			var marble_color = marble.get_color()
			remove_child(marble)
			marble.queue_free()
			var magic = marble_animation.instantiate()
			add_child(magic)
			var marble_animation_skin = magic.get_child(0) #Skin
			var marble_animation_effect = magic.get_child(1) #Effect
			marble_animation_effect.position = initial_position
			marble_animation_skin.set_color_by_name(marble_color)
			var animatron = marble_animation_skin.get_child(0)
			var anime = animatron.get_animation("disappearing")
			anime.track_set_key_value(0,0, initial_position)
			animatron.play("disappearing")
			await animatron.animation_finished

func cleanup_all():
	# TODO: I think this code does nothing. Clean up
	for node in get_children():
		if node.is_class("Gate") or \
		  node.is_class("Marble"):
			remove_child(node)
			node.queue_free()


func _on_marble_jam() -> void:
	if jam_count == 0:
		marbles_to_handle = []
		for node in $Board/BackBoard.get_children():
			if node.name.begins_with("GateSection"):
				if node.marbles.size() > 1:
					for marble in node.marbles:
						marbles_to_handle.append(marble)
						
	if marbles_to_handle.size() == 0:
		return
		
	jam_count += 1
	var marble = marbles_to_handle[0]
	var handle_marble_scn = preload('res://src/handle_the_marble.tscn').instantiate()
	add_child(handle_marble_scn)
	handle_marble_scn.marble_info(marble)
	marbles_to_handle.erase(marble)
		
	


func _on_victory_button_pressed() -> void:
	get_tree().change_scene_to_file("res://src/thanks.tscn")

func from_step21_to_step22():
	step21_complete = true
	if step22_complete == false:
		$MainAlphaRect/RichTextLabel1.visible = false
		$MainAlphaRect/RichTextLabel2.visible = true
		$AlphaRect1.visible = false
		$AlphaRect2.visible = false
		$MainAlphaRect/TextureRect.visible = false
		$MainAlphaRect/TextureRect2.visible = false
		$MainAlphaRect/TextureRect4.visible = true
	
func from_step22_to_step23():
	step22_complete = true
	$MainAlphaRect/TextureRect4.visible = false
	$MainAlphaRect/RichTextLabel2.visible = false
	$MainAlphaRect/RichTextLabel3.visible = true
	$MainAlphaRect/TextureRect3.visible = true

func from_step23_to_step24():
	step23_complete = true
	$TextBG.visible = true
	$TextBG/RichTextLabel.visible = true
	$NextStep.visible = true
	
func _on_next_step_pressed() -> void:
	var source = "res://src/tuto/step5.tscn"
	get_tree().change_scene_to_file.bind(source).call_deferred()
	
func update_lang():
	if $"/root/SettingsData".lang == "es":
		$TextBG/RichTextLabel.text = text4_es
	elif $"/root/SettingsData".lang == "fr":
		$TextBG/RichTextLabel.text = text4_fr
	else:
		$TextBG/RichTextLabel.text = text4_en
			

var text4_es = """ [center] Si la canica lanzada no llega hasta abajo, ésta se queda en el tablero.
[/center]
"""

var text4_en = """ [center] If the marble doesn't reach the bottom, it stays on the board.
[/center]
"""


var text4_fr = """ [center] Si la bille lancée n'atteint pas le socle, elle reste dans le plateau.
[/center]
"""
